/*
   Copyright (C) 1988, 1989, 1990 The GAP Development Company

   All Rights Reserved
   
   This source code is considered to be confidential information proprietary
   to The GAP Development Company.  It is prohibited and punishable by law
   to use or duplicate any part of this information without express
   written consent from GAP Development Company.

   This source code is being provided for informational purposes only.


   UTIIMPRT.C

   UTI module that imports echo mail from a UTI file into the ISAM
   message base.


   To compile : cl /c /FPa utiimprt.c

   To link    : link utiimprt+obj\ctclibss,,NUL.MAP,+ctreess
              : link utiimprt+obj\ctclibsm,,NUL.MAP,+ctreesm  // multi user

*/

#define LINT_ARGS

#define MULTIUSER
//#define SINGLEUSER

#pragma check_stack(off)
#pragma pack(1)

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <time.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
#include <share.h>
#include <sys\locking.h>

#pragma pack()                            // cannot pack Ctree structs
#include <ctstdr.h>
#include <ctifil.h>
#pragma pack(1)

#pragma intrinsic(memset)

   /*********************************************************************/
   /*                   FILE MODES FOR PARAMETER FILE                   */
   /*                                                                   */
   /*                                 Multi     Single                  */
   /*    Messages:                                                      */
   /*                Filemode1          5         4            DAT      */
   /*                Filemode2          1         0            IDX      */
   /*                                                                   */
   /*    Users:                                                         */
   /*                Filemode3          3         2            DAT      */
   /*                Filemode4          3         2            IDX      */
   /*********************************************************************/

#ifdef MULTIUSER                          // multi user file mode parameters
  #define filemode1   '5'
  #define filemode2   '1'
  #define filemode3   '3'
  #define filemode4   '3'
  #define incnum       1                  // used to increment file numbers
#else                                     // single user file mode parameters
  #define filemode1   '4'
  #define filemode2   '0'
  #define filemode3   '2'
  #define filemode4   '2'
  #define incnum       0                  // used to increment file numbers
#endif                                    // multi-user needs dummylok.dat file

   int datno     = 0+incnum;              // data file number to use     
   int idx4      = 4+incnum;              // index file number to use    
   int forumnum  = 0;                     // forum number to work with   
   long active   = 0L;                    // number of active messages   
   long highnum  = 0L;                    // high message # for MESSNO.DAT   
   long rlymesno = 0L;                    // last # read from UTI.TXT
   long mesno    = 0L;                    // last # read from MESSNO.DAT

   char ourmail    [60];                  // path to our MSGS.xxx files  
   char hismail    [60];                  // path to UTI MAIL.TXT file   
   char dflt_path  [60];                  // path to GAP default dir
   char messpath   [60];                  // path to MESSNO.DAT
   char anystring [250];                  // global garbage collector

   unsigned char *buffer = NULL;          // pointer to message buffer

   struct MAIL                            // structure of the header rec 
      {
      char del;                           // delete flag                 
      char to           [32];             // left to whom (indexed)      
      char from         [32];             // left by whom (indexed)      
      char sub          [26];             // subject (thread - indexed)  
      long num;                           // message number (indexed)    
      long date;                          // date left                   
      long readdate;                      // date when it was read       
      long refer;                         // refer to message number     
      long activemess;                    // number of active messages   
      char read;                          // has it been read by "to"?   
      char type;                          // private, passwrd,open, etc  
      char passwd       [13];             // password if protected       
      } mail,ref_mail;                    // 126 bytes 

   struct FORUM
      {
      char name         [21];             // forum name                   
      char directory    [31];             // forum dir path               
      char gen_dir      [31];             // forum gen dir path           
      char up_dir       [31];             // upload dir path              
      char private_up    [2];             // keep uploads private?        
      char keep_up       [2];             // keep uploads in forum?       
      char auto_join     [2];             // auto rejoin forum?           
      char handle        [2];             // allow handles?               
      char priv_mail     [2];             // make all mail private?       
      char pub_mail      [2];             // make all mail public?        
      char subscrip      [9];             // expired subscrip date        
      char net_mail      [2];             // is this a netmail forum?     
      char sysop1       [31];             // # 1 sysop this forum         
      char sysop2       [31];             // # 2 sysop this forum         
      char main_dirs     [2];             // include main board dirs?     
      int bullets;                        // number of bulletins          
      int dirs;                           // number of directories        
      int security;                       // additional security          
      int time;                           // additional time              
      int sys1_level;                     // level for sysop 1            
      int sys2_level;                     // level for sysop 2            
      } forum;                            // 213 bytes                    


   /**********************************************************************/
   /* A shortened version of GAPBBS.CNF.  All we need is the path to the */ 
   /* main directory and the name of the forum.dat file                  */ 
   /**********************************************************************/

   struct GAP_CONFIG
      {
      char sysname       [16];             // sysop name                   
      char syspas        [13];             // sysop local password         
      int sys_read;                        // level to read private mail   
      int sys_edit;                        // level to edit msg headers    
      int overwrite;                       // level to overwrite uploads   
      int sys_level1;                      // level to view callers log    
      int sys_level2;                      // level to view users          
      int sys_level3;                      // level to view a file         
      int sys_level4;                      // level to maintain user file  
      int sys_level5;                      // level to exit to dos         
      int sys_level5a;                     // level to execute dos funcs   
      int sys_level6;                      // level to view other nodes    
      int sys_level6a;                     // level to log off other node  
      int sys_level7;                      // level to recover messages    

      char answer         [8];             // questionnaire answers file   
      char caller         [7];             // caller file                  
      char doors          [9];             // doors.dat file               
      char forumdat       [9];             // name of forum (.DAT) file    
      char entry_news     [8];             // news file                    
      char exit_news      [8];             // exit news file               
      char pwrd           [9];             // time / security file         
      char fsec           [9];             // files security file          
      char msgs           [9];             // messages file                
      char new_welcome    [8];             // new user welcome file        
      char answer_regis   [8];             // not used                     
      char remote         [9];             // remote.sys file              
      char tcan           [9];             // trashcan file                
      char users          [9];             // users file                   
      char welcome        [8];             // welcome file                 

      char uploads        [26];            // upload directory             
      char gen            [26];            // general directory            
      char maindir        [26];            // main directory               
      char help           [26];            // help directory               
      char loose_dir      [26];            // directory where loose files are

      char dumb_modem      [2];            // use true ring detect?        
      char modem_init     [45];            // initilization string         
      char modem_off      [45];            // off hook string              
      int  modem_wait;                     // seconds to wait for carrier  
      int  connect_wait;                   // seconds to wait for connect  
      char modem_reset    [45];            // reset string                 
      char port            [2];            // com port                     
      char speed           [6];            // speed to open modem at       
      char baud_300        [2];            // restrict Baud rates?         
      char baud_start      [6];            // restrict start time          
      char baud_end        [6];            // restrict end time            
      char baud_rest      [10];            // highest baud rate to restrict
      char lock_baud       [2];            // lock in the baud rate        
      char zstream         [2];            // full flow zmodem uploads     

      int menuA;                           //  main menu security levels   
      int menuB;
      int menuC;
      int menuD;
      int menuE;
      int menuF;
      int menuG;
      int menuH;
      int menuI;
      int menuJ;
      int menuK;
      int menuL;
      int menuM;
      int menuN;
      int menuO;
      int menuP;
      int menuQ;
      int menuR;
      int menuS;
      int menuT;
      int menuU;
      int menuV;
      int menuW;
      int menuX;
      int menuY;
      int menuZ;

      int menu_chat;                      // chat with other nodes        
      int menu_open;                      // open a door                  
      int menu_who;                       // who's on other nodes         
      int regis;                          // registration level           
      int year_regis;                     // level for subscrip expired   

      int viewG;                          //  view/change stats           
      int viewN;              
      int viewL;              
      int viewT;             
      int viewX;              

      int fileA;                          //  file menu security levels   
      int fileB;                          // browse text files            
      int fileC;
      int fileD;
      int fileE;
      int fileF;                          //  file menu security levels   
      int fileG;
      int fileH;
      int fileI;
      int fileJ;
      int fileK;
      int fileL;
      int fileM;
      int fileN;
      int fileO;
      int fileP;
      int fileQ;
      int fileR;
      int fileS;
      int fileT;
      int fileU;
      int fileV;
      int fileW;
      int fileX;
      int fileY;
      int fileZ;

      int blts;                            // number of main bulletins     
      int cnfs;                            // number of FORUMS             
      int dirs;                            // number of directories        
      int printer;                         // printer port                 
      char time_limit     [2];             // enforce daily time limit     
      char byte_cnt       [2];             // enforce daily byte count     
      char private_up     [2];             // use private upload directory 
      char dir_up         [2];             // allow directory uploads      
      char closed         [2];             // closed board flag            
      char handle         [2];             // allow handles                
      char chg_name       [2];             // allow name changes           
      char regis_chk      [2];             // edit registration info       
      char subscribe      [2];             // enforce yearly registration  
      int sub_days;                        // # of days in subscrip period 
      char quick          [2];             // allow quick logons           
      char priv_mail      [2];             // make all mail private?       

      char board          [51];            // board name                   
      char comment        [51];            // comment prompt               
      char twit           [51];            // twit string with F8          
      char lockout        [51];            // lock out string with F2      
      char nu1            [51];            // future implementation        
      char nu2            [51];            // future implementation        
      int  node;                           // node number                  
      char events         [2];             // use timed events             
      char event1         [6];             // event 1 time                 
      char event2         [6];             // event 2 time                 
      int  color;                          // default color 1 - 15         
      int  free_space;                     // minimum free upload space    
      int  max_msgs;                       // maximum number of messages   
      int  maxlines;                       // maximum lines per messae     
      }config;

   struct MESSNO
      {  
      long status;                        // multiuser status flags      
      long highmain;                      // highest used message number 
      long fhigh [90];                    // for main, and each of the   
      };                                  // forums.                     


   /**********************************************************************/
   /*  Parameters for the ISAM mail files.  The path and name to the     */ 
   /*  message file will be filled in at run time.                       */ 
   /**********************************************************************/

#ifdef MULTIUSER
   IFIL dummylock =
      {
      anystring,0,128,0,3,0,0,0,0,"delflag","delflag"
      };
#endif

   ISEG segments[] =
      {
      {1, 32,0},
      {33,32,0},
      {65,26,0},
      {91, 4,0}
      };

   IIDX indices[] =
      {
      {36,0,1,1,32,1,&segments[0],"to_idx"},
      {36,0,1,1,32,1,&segments[1],"from_idx"},
      {30,0,1,1,32,1,&segments[2],"sub_idx"},
      { 4,1,0,0, 0,1,&segments[3],"num_idx"}
      };

   IFIL imail =
      {
      ourmail,0+incnum,126,4096,filemode1,4,4096,filemode2,indices,"deleted","text"
      };

   extern int isam_err, isam_fil;

   extern time_t _dtoxtime(int yr,int mo,int dy, int hr, int mn, int sc);
   static  int pascal read_net(void );
   static  int pascal open_gapcnf(void );
   static  int pascal open_forumdat(int which);
   static  char *trim(char *str);
   static  void pascal update_high(void );
   static  long pascal message_num(int how);
   static  void pascal write_descp(char *path);
   extern  void pascal run_menu(int crap);
   extern  void pascal log_call(int crap,char *crap1);


void main(argc,argv)
int argc;
char *argv[];
{
   char descrep;                          // if 1, write descrepancy file
   int error = 0;

   strcpy(hismail,argv[2]);               // path to UTI file                  
   forumnum = atoi(argv[1]);              // forum to import messages to

   // /NETWORK can be either ARG 3 or 4 so check 3 and if not /NETWORK
   // then assume it is the path to descrepancy file

   if (argc > 3 && (strcmp(argv[3],"/NETWORK")) != 0)
      descrep = 1;                        // write descrepancy file
   else
      descrep = 0;                        // dont write it     

   if ( (error = open_gapcnf()) != 0)     // Open UTI configuration, get
      exit (error);                       // path to GAPBBS.CNF

   if ( (error = open_forumdat(forumnum)) != 0)   // Open and read data for
      exit (error);                               // this Forum

   if (intisam(12,6,4))                   // get some memory for index buffers 
      exit(isam_err);

#ifdef MULTIUSER
   sprintf(anystring,"%s\\DUMMYLOK",config.maindir);

   if (opnifil(&dummylock))               // open DUMMYLOK.DAT
      exit(isam_err);
#endif

   if (opnifil(&imail))                   // open the ISAM files               
      exit(isam_err);

   error = read_net();                    // read UTI file

   if (error == -2)                       // -2 = cant allocate memory
      {
      cputs("\r\n\r\nCannot Allocate Memory For Message Buffer!\r\n");
      }

   clisam();                              // close the ISAM files


   // if mesno != rlymesno (read from UTI) then someone saved
   // a message while we are processing this Forum

#ifdef MULTIUSER
   if (!error && descrep && mesno != rlymesno)
       write_descp(argv[3]);               // write difference in #'s to file
#else
   if (!error)                             // dont touch messno.dat if error
      update_high();                       // update messno with high number
#endif

   exit(error);
}

   /******************************************************************/
   /*  read the network messages file.                               */ 
   /*  return 0 if successful, errno if error                        */ 
   /******************************************************************/

static int pascal read_net()
{
   int r;
   int error = 0;
   int yr,mo,dy,hr,mn;
   int count;
   unsigned numbytes;
   unsigned char *ptr;

   FILE *fp;

   if ( (fp = fopen(hismail,"rt")) == NULL)
      return (-1);

   while ( (buffer = (unsigned char *) malloc((((config.maxlines+5)*73)+sizeof(mail)))) == NULL
            && config.maxlines > 0)
        --config.maxlines;

   if (buffer == NULL)                    // error allocating memory
      return(-2);
   
   if (config.maxlines <= 0)              // not enough lines left
      {
      free(buffer);
      return(-2);
      }

   /*****************************************************************/
   /*  Need to first read the last message and get the number of    */ 
   /*  active messages.                                             */ 
   /*****************************************************************/

#ifndef MULTIUSER
   if (lstrec(idx4,(char *) &mail))       // there are no messages
      active = 1L;
   else
      active = mail.activemess + 1L;      // increment active by 1
#endif

   while ( !feof(fp) )                    // read till end of file
      {
      memset(&mail,'\0',sizeof(mail));    // NULL out header

      mail.del  = 0x10;                   // flag that mes is from net

      if (fgets(mail.to,32,fp) == NULL)   // who the message is for    
         break;                           // if NULL, end of file      
      r = strlen(mail.to);
      mail.to[r-1] = 0;                   // get rid of NL             

      fgets(mail.from,32,fp);             // who the message is from   
      r = strlen(mail.from);
      mail.from[r-1] = 0;                 // get rid of NL             

      fgets(anystring,99,fp);             // subject of message        
      r = strlen(anystring);              // cant be longer than 25    
      if (r > 25)
         anystring[25] = 0;
      else
         anystring[r-1] = 0;
      strcpy(mail.sub,anystring);

#ifdef MULTIUSER
      fscanf(fp,"%ld",&rlymesno);         // message number
      mail.num = rlymesno;
#else
      fscanf(fp,"%ld",&mail.num);         // message number
      mail.activemess = active++;         // keep track of active msgs
#endif

      fgets(anystring,99,fp);             // get CR/LF                 
      fgets(anystring,99,fp);             // refer to # or "Relay"     

      mail.refer = atol(anystring);       // if number convert it      

      fgets(anystring,99,fp);             // get the date              
      anystring[2] = '/';
      anystring[5] = '/';
      anystring[8] = 0;
      sscanf(anystring,"%d/%d/%d",&mo,&dy,&yr);   // convert it      

      fscanf(fp,"%d:%d",&hr,&mn);         // get time                  

      /**************************************************************/
      /* Internal MSoft function to convert a DOS date into a UNIX  */ 
      /* long number. This will have to be changed in the year 2000 */ 
      /**************************************************************/

      mail.date = _dtoxtime((yr+1900)-1980,mo,dy,hr,mn,0);

      fgets(anystring,99,fp);             // get the CR/LF             

      fgets(anystring,99,fp);             // security
      if ((strncmp(anystring,"PRIVATE",7)) == 0)
         mail.type = 'R';

      fgets(anystring,99,fp);             // get 'Read' flag
      if (anystring[0] == 'Y')            // message has been read
         {
         mail.read = 'Y';
         mail.readdate = mail.date;
         }
      else
         mail.read = 'N';                 // message has not been read

      while ( !feof(fp) )                 // gobble up till 'TEXT:' found
         {
         if (fgets(anystring,99,fp) == NULL)
            break;
         if ((strncmp(anystring,"TEXT",4)) == 0)
            break;
         }

      numbytes = sizeof(mail);            // keep track of # of bytes  
      ptr = buffer + numbytes;            // position ptr into buffer  

      count = 0;                          // count # of lines

      while ( !feof(fp) )                 // read the message text     
         {
         *(ptr) = fgetc(fp);              // get the character         
         if(*(ptr) == 0xFF)               // end of this message       
            {
            r = fgetc(fp);                // get the NL or we will end 
            break;                        // up saving blank message   
            }
         if(*(ptr) == '\n')               // GAP end of line delimiter 
            {
            *(ptr) = 0;
            ++count;                      // count # of lines
            }

         if (count <= config.maxlines)    // dont overflow buffer
            {
            ++ptr;                        // pointer into buffer
            ++numbytes;                   // number of bytes read
            }
         }

      /**************************************************************/
      /*  Need to first check to see if the "refer to" message      */ 
      /*  exists.  If not, set refer to = 0.                        */ 
      /**************************************************************/

      if (mail.refer != 0L)
         {
         error = eqlrec(idx4,(char *) &mail.refer,(char *) &ref_mail);

         if (error || (!error && (ref_mail.del & 0x80) ))
            mail.refer = 0L;
         }


      /**************************************************************/
      /*  Add the file into the ISAM message base.  There should be */ 
      /*  no errors, but if there are, then the most likely cause   */ 
      /*  is a duplicate message number.  If that is the case, then */ 
      /*  all the rest of the echo messages are going to cause an   */ 
      /*  error, so no since continuing.                            */ 
      /**************************************************************/

      /*********************************************************/
      /*  Lock MESSNO.DAT and keep it locked until we save     */
      /*  this message.  Tries up to 10 times to get a lock    */
      /*  If we cant get a lock, no since continuing since     */
      /*  something is wrong.                                  */
      /*********************************************************/

#ifdef MULTIUSER
      mesno = message_num(0);             // get a message number

      if (mesno < 1L)                     // error opening or locking file
         break;

      if ((lkisam(ENABLE)))               // cant lock message file
         {
         message_num(1);                  // unlock MESSNO.DAT
         break;
         }

      mail.num = mesno;                   // assign message number

      if (lstrec(idx4,(char *) (char *) &ref_mail))  // there are no messages
         mail.activemess = 1L;
      else
         mail.activemess = ref_mail.activemess + 1L; // increment active by 1
#endif

      memmove(buffer,&mail,sizeof(mail)); // copy header to buffer

      if ((error = addvrec(datno,buffer,numbytes)) != 0)
         break;

#ifdef MULTIUSER
      message_num(1);                     // unlock MESSNO.DAT
      lkisam(FREE);                       // unlock ISAM file
#else
      highnum = mail.num;                 // keep track of high number 
#endif
      }
   
   fclose(fp);
   free(buffer);                          // free memory
   return (error);
}


   /******************************************************************/
   /* Open GAPBBS.CNF and read only as much of the record as we need */ 
   /******************************************************************/

static int pascal open_gapcnf()
{
   int gapfile;

   FILE *fp;

   if ( (fp = fopen("GAPUTI.CNF","rt")) == NULL)
      return (errno);

   fscanf(fp,"%s",dflt_path);             // path to GAP default directory     
   fclose(fp);

   strcat(dflt_path,"\\GAPBBS.CNF");

#ifdef MULTIUSER
   gapfile = sopen(dflt_path,O_RDONLY | O_BINARY,SH_DENYNO);
#else
   gapfile = open(dflt_path,O_RDONLY | O_BINARY);
#endif

   if(gapfile == -1)
      return (errno);

   read(gapfile,(char *) &config,sizeof(config));
   close(gapfile);

   trim(config.maindir);                  // get rid of trailing spaces       
   trim(config.forumdat);
   trim(config.msgs);

   sprintf(messpath,"%s\\MESSNO.DAT",config.maindir);

   return (0);
}

static int pascal open_forumdat(which)    // open and read the forum dat file 
int which;                                // which record to read              
{
   int forumfile;
   char forumdir  [60];

   if (which)                             // if this isnt the main board
      {
      sprintf(forumdir,"%s\\%s.DAT",config.maindir,config.forumdat);

#ifdef MULTIUSER
      forumfile = sopen(forumdir, O_RDONLY | O_BINARY,SH_DENYNO);
#else
      forumfile = open(forumdir, O_RDONLY | O_BINARY);
#endif

      if (forumfile == -1)
         return (errno);

      lseek(forumfile,(long) (sizeof(forum) * (which -1)),SEEK_SET);
      read(forumfile,(char *) &forum,sizeof(forum));
      close(forumfile);

      trim(forum.directory);
      sprintf(ourmail,"%s\\%s",forum.directory,config.msgs);
      }
   else
      sprintf(ourmail,"%s\\%s",config.maindir,config.msgs);

   return(0);
}

static char *trim(str)                    // removes trailing spaces          
char *str;
{
   char *ptr;
   
   ptr = str + strlen(str) -1;
  
   while(ptr >= str && *ptr == ' ')
      *ptr-- = 0;

   return(str);
}


   /*******************************************************************/
   /*  Update the high message number in MESSNO.DAT                   */ 
   /*  Used only by single user systems                               */
   /*******************************************************************/

static void pascal update_high() 
{
   struct MESSNO messno;
   int fp;

   fp = open(messpath,O_RDWR | O_BINARY);

   if (fp != -1)
      {
      read(fp,(char *) &messno,sizeof(messno));
      lseek(fp,0L,SEEK_SET);

      messno.fhigh[forumnum-1] = highnum;

      write(fp,(char *) &messno,sizeof(messno));
      close(fp);
      }
}

     /******************************************************************/
     /*  Locks messno.dat.  Return -1L if unsuccessful, or a message   */
     /*  number if ok.                                                 */
     /******************************************************************/


static long pascal message_num(how)
int how;                                  // 0 = lock and get #, 1 = unlock
{
   register bobo;
   register r;
   int ok;
   struct MESSNO messno;
   static int fp = -1;
   long messagenum;

   if (!how)                              // lock and get number
      {
      fp = sopen(messpath,O_RDWR | O_BINARY,SH_DENYNO);

      if (fp == -1)                       // cant open file
         {
         return (-1L);
         }

      /**************************************************************/
      /*  If multi-user, lock the file & keep it locked.            */
      /*  File stays locked until after the message is saved.       */
      /*  We will let Msoft handle the lock tries                   */
      /**************************************************************/

      if ((locking(fp,LK_LOCK,(long) sizeof(messno))) == -1)
         {
         close(fp);
         return (-1L);
         }

      read(fp,(char *) &messno,sizeof(messno));
      lseek(fp,0L,SEEK_SET);

      /**************************************************************/
      /*    First, get the message number and increment it by one.  */
      /**************************************************************/

      ++messno.fhigh[forumnum-1];         // increment the number
      messagenum = messno.fhigh[forumnum-1];

      write(fp,(char *) &messno,sizeof(messno));
      return(messagenum);                 // return with the message number
      }

   // unlock file but only if we were successful in locking it

   if (how && fp != -1)                   // unlock file and close it
      {
      lseek(fp,0L,SEEK_SET);
      locking(fp,LK_UNLCK,(long) sizeof(messno));
      close(fp);
      fp = -1;
      return(1L);
      }
}

static void pascal write_descp(path)
char path[];
{
   FILE *fp;

   fp = fopen(path,"w+t");

   if (fp != NULL)
      {
      fprintf(fp,"%ld\n",mesno-rlymesno);
      fclose(fp);
      }
}


   /********************************************************************/
   /*  These are two dummy functions for the ISAM routines.  If there  */ 
   /*  is a fatal error, the fatal is normally logged to the GAP       */ 
   /*  caller log and then the GAP exit routine is run.  If this       */ 
   /*  program ever exits with an error code of 99, then there was a   */ 
   /*  a fatal ISAM error and the sysop will need to rebuild his ISAM  */ 
   /*  data files.                                                     */ 
   /*  They need to be declared because the ISAM routines expect them  */ 
   /*  to be present.                                                  */ 
   /********************************************************************/

void pascal run_menu(crap)
int crap;
{
   clisam();
   exit(99);
}

void pascal log_call(crap,crap1)
int crap;
char crap1[];
{
}

